﻿// Fontx.c : 
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	FONT_FILE_ZEN_12_12		"_FONT/DRUG12Z1.FEF"
#define	FONT_FILE_HAN_6_12		"_FONT/DRUG12HN.FEF"

typedef struct _FontMetricsBlock {
	unsigned short	StartCode;		// 開始コード(SJIS文字コード)
	unsigned short	EndCode;		// 終了コード(SJIS文字コード)
	int				Num;			// コード数
	//unsigned char	*FontData;		// フォントデータ
	unsigned short	*FontData;		// フォントデータ
} FONTMETRICSBLOCK;

typedef struct _FontMetrics {
	int		Width;		// フォント幅(dot)
	int		Height;		// フォント高さ(dot)
	int		BytesWidth;			// フォントイメージの横バイト数
	int		BytesPerChar;		// 1文字当たりバイト数(フォントデータ)
//	int		BytesPerCharWrk;	// 1文字当たりバイト数(変換ワークデータ)
	int		WordsPerCharWrk;	// 1文字当たりワード数(変換ワークデータ)
	int		Code;		// 文字コード
	int		Blocks;		// コードブロック数
	FONTMETRICSBLOCK	*Block;
} FONTMETRICS;

static FONTMETRICS		FontZen_12_12;
static FONTMETRICS		FontHan_6_12;


void TransAxisBytes(unsigned char *from, unsigned char *to);
//unsigned char *GetFontData(FONTMETRICS *Font, unsigned short sjis_code);
unsigned short *GetFontData(FONTMETRICS *Font, unsigned short sjis_code);

/*
 *	フォントファイルを読み取り、FONTMETRICSデータを作成する。
 */
int GetFontMetrics(char *FontFile, FONTMETRICS *FontMetrics){
	FILE	*fp;
	unsigned char	buff[128];
	unsigned char	work1[8], work[2];
	int		i, j, k;
	int		idxBlock, idxFontInBlock;

	if ((fp = fopen(FontFile, "r")) == NULL){
		printf("フォントファイルオープンエラー\n");
		return 0;
	}
	fread(buff, 17, 1, fp);		// 
	if (strncmp((char *)buff, "FONTX2", 6) != 0){
	//if (_tcsncmp((char *)buff, _T("FONTX2"), 6) != 0){
		printf("FONTXでない\n");
		return 0;
	}
	FontMetrics->Width = buff[14];
	FontMetrics->Height = buff[15];
	FontMetrics->Code = buff[16];
	if (FontMetrics->Code == 0){	// ANK
		FontMetrics->Blocks = 1;
		FontMetrics->BytesWidth = 1;
//		FontMetrics->BytesPerCharWrk = 16;
		FontMetrics->WordsPerCharWrk = 8;
		FontMetrics->BytesPerChar = FontMetrics->BytesWidth * FontMetrics->Height;
		FontMetrics->Block = (FONTMETRICSBLOCK *)malloc(FontMetrics->Blocks * sizeof(FONTMETRICSBLOCK));

		i = 0;
		FontMetrics->Block[i].StartCode = 0;
		FontMetrics->Block[i].EndCode = 0xFF;
		FontMetrics->Block[i].Num = FontMetrics->Block[i].EndCode - FontMetrics->Block[i].StartCode + 1;
		//FontMetrics->Block[i].FontData = (unsigned char *)malloc(FontMetrics->Block[i].Num * FontMetrics->BytesPerCharWrk);
		FontMetrics->Block[i].FontData = (unsigned short *)malloc(FontMetrics->Block[i].Num * FontMetrics->WordsPerCharWrk * 2);

		for (j=0; j<FontMetrics->Block[i].Num; j++){
			unsigned char	to1[8], to2[8];
			unsigned short	*to_p;
			
			fread(buff, FontMetrics->BytesPerChar, 1, fp);

			// 上ブロック
			for (k=0; k<8; k++){
				work1[k] = buff[k];
			}
			//TransAxisBytes(work1, FontMetrics->Block[i].FontData + (j * FontMetrics->BytesPerCharWrk));
			TransAxisBytes(work1, to1);

			if (FontMetrics->Height > 8){
				// 下ブロック
				for (k=0; k<8; k++){
					if (k < (FontMetrics->Height - 8)){
						work1[k] = buff[k + 8];
					} else {
						work1[k] = 0;
					}
				}
				//TransAxisBytes(work1, FontMetrics->Block[i].FontData + (j * FontMetrics->BytesPerCharWrk) + 8);
				TransAxisBytes(work1, to2);
				
//				to_p = FontMetrics->Block[i].FontData + (j * FontMetrics->BytesPerCharWrk);
				to_p = FontMetrics->Block[i].FontData + (j * FontMetrics->WordsPerCharWrk);
				for (k=0; k<8; k++){
					*to_p++ = (to1[k] | (((unsigned short)to2[k]) << 8));
				}
			}
		}

	} else {
		fread(buff, 1, 1, fp);			// コードブロック数(NB)
		FontMetrics->Blocks = buff[0];
		FontMetrics->BytesWidth = ((FontMetrics->Width + 7) / 8);
//		FontMetrics->BytesPerCharWrk = 32;	// 12x12, 16x16フォント(FontDataバッファ中の１文字当たりバイト数)
		FontMetrics->WordsPerCharWrk = 16;	// 12x12, 16x16フォント(FontDataバッファ中の１文字当たりワード数)
		FontMetrics->BytesPerChar = FontMetrics->BytesWidth * FontMetrics->Height;
		FontMetrics->Block = (FONTMETRICSBLOCK *)malloc(FontMetrics->Blocks * sizeof(FONTMETRICSBLOCK));

		//printf(_T("FontMetrics Width=%d\n"), FontMetrics->Width);
		//printf(_T("FontMetrics BytesPerChar=%d\n"), FontMetrics->BytesPerChar);

		// コードブロックテーブル
		for (i=0; i<FontMetrics->Blocks; i++){
			fread(buff, 4, 1, fp);		// 各ブロックの開始コードと終了コード

			FontMetrics->Block[i].StartCode = (buff[0] | ((unsigned short)buff[1] << 8));		// ブロックの開始コード
			FontMetrics->Block[i].EndCode = (buff[2] | ((unsigned short)buff[3] << 8));			// 終了コード
			FontMetrics->Block[i].Num = FontMetrics->Block[i].EndCode - FontMetrics->Block[i].StartCode + 1;	// ブロック内の文字数
			//FontMetrics->Block[i].FontData = (unsigned char *)malloc(FontMetrics->Block[i].Num * FontMetrics->BytesPerCharWrk);
			FontMetrics->Block[i].FontData = (unsigned short *)malloc(FontMetrics->Block[i].Num * FontMetrics->WordsPerCharWrk * 2);
			//printf(_T("FontMetrics StartCode=%d num=%d\n"), FontMetrics->Block[i].StartCode, FontMetrics->Block[i].Num);
		}

		// フォントイメージ
		for (i=0; i<FontMetrics->Blocks; i++){
			for (j=0; j<FontMetrics->Block[i].Num; j++){
				fread(buff, FontMetrics->BytesPerChar, 1, fp);		// 1文字分のフォントイメージ

				// 16x16フォントを１ブロック=8x8ドットの４ブロックに分割。
				// 
				// ABCDEFGH
				// IJKLMNOP
				// QRSTUVWX
				// abcdefgh
				// ijklmnop
				// qrstuvwx
				// 01234567
				// ｱｲｳｴｵｶｷｸ
				// 
				// フォントファイル中では ABCDEFGH, IJKLMNOP, ... で各バイトを構成している。これを
				// ｱ0qiaQIA, ｲ1rjbRJB, ... というバイト構成に変換する。ｱ,ｲ,...が上位ビットとなる。

				if (FontMetrics->BytesWidth == 2){
					unsigned char	to1[8], to2[8], to3[8], to4[8];
					unsigned short	*to_p;
					// 左上ブロック
					for (k=0; k<8; k++){
						work1[k] = buff[2*k];
					}
					//TransAxisBytes(work1, FontMetrics->Block[i].FontData + (j * FontMetrics->BytesPerCharWrk));
					TransAxisBytes(work1, to1);
					// 右上ブロック
					for (k=0; k<8; k++){
						work1[k] = buff[2*k + 1];
					}
					//TransAxisBytes(work1, FontMetrics->Block[i].FontData + (j * FontMetrics->BytesPerCharWrk) + 8);
					TransAxisBytes(work1, to2);
					
					if (FontMetrics->Height > 8){
						// 左下ブロック
						for (k=0; k<8; k++){	//	12x12フォントでは4
							if (k < (FontMetrics->Height - 8)){
								work1[k] = buff[16 + 2*k];
							} else {
								work1[k] = 0;
							}
						}
						//TransAxisBytes(work1, FontMetrics->Block[i].FontData + (j * FontMetrics->BytesPerCharWrk) + 16);
						TransAxisBytes(work1, to3);
						// 右下ブロック
						for (k=0; k<8; k++){	//	12x12フォントでは4
							if (k < (FontMetrics->Height - 8)){
								work1[k] = buff[16 + 2*k + 1];
							} else {
								work1[k] = 0;
							}
						}
						//TransAxisBytes(work1, FontMetrics->Block[i].FontData + (j * FontMetrics->BytesPerCharWrk) + 24);
						TransAxisBytes(work1, to4);
						
						// 
//						to_p = FontMetrics->Block[i].FontData + (j * FontMetrics->BytesPerCharWrk);
						to_p = FontMetrics->Block[i].FontData + (j * FontMetrics->WordsPerCharWrk);
						for (k=0; k<8; k++){
							*to_p++ = (to1[k] | (((unsigned short)to3[k]) << 8));
						}
						for (k=0; k<8; k++){
							*to_p++ = (to2[k] | (((unsigned short)to4[k]) << 8));
						}
					}
				}

				//printf(_T("FontMetrics i=%d j=%d\n"), i, j);
			}
		}
	}

	fclose(fp);

	return 1;
}


/*
 *	長さ８のバイト配列のビット位置を縦横入れ替え
 */
void TransAxisBytes(unsigned char *from, unsigned char *to){
	int		i, j;

	for (i=0; i<8; i++) to[i] = 0;
	
	to[0] |= ((from[0] & 0x80) >> 7);
	to[1] |= ((from[0] & 0x40) >> 6);
	to[2] |= ((from[0] & 0x20) >> 5);
	to[3] |= ((from[0] & 0x10) >> 4);
	to[4] |= ((from[0] & 0x08) >> 3);
	to[5] |= ((from[0] & 0x04) >> 2);
	to[6] |= ((from[0] & 0x02) >> 1);
	to[7] |= ((from[0] & 0x01) >> 0);

	to[0] |= ((from[1] & 0x80) >> 6);
	to[1] |= ((from[1] & 0x40) >> 5);
	to[2] |= ((from[1] & 0x20) >> 4);
	to[3] |= ((from[1] & 0x10) >> 3);
	to[4] |= ((from[1] & 0x08) >> 2);
	to[5] |= ((from[1] & 0x04) >> 1);
	to[6] |= ((from[1] & 0x02) >> 0);
	to[7] |= ((from[1] & 0x01) << 1);

	to[0] |= ((from[2] & 0x80) >> 5);
	to[1] |= ((from[2] & 0x40) >> 4);
	to[2] |= ((from[2] & 0x20) >> 3);
	to[3] |= ((from[2] & 0x10) >> 2);
	to[4] |= ((from[2] & 0x08) >> 1);
	to[5] |= ((from[2] & 0x04) >> 0);
	to[6] |= ((from[2] & 0x02) << 1);
	to[7] |= ((from[2] & 0x01) << 2);

	to[0] |= ((from[3] & 0x80) >> 4);
	to[1] |= ((from[3] & 0x40) >> 3);
	to[2] |= ((from[3] & 0x20) >> 2);
	to[3] |= ((from[3] & 0x10) >> 1);
	to[4] |= ((from[3] & 0x08) >> 0);
	to[5] |= ((from[3] & 0x04) << 1);
	to[6] |= ((from[3] & 0x02) << 2);
	to[7] |= ((from[3] & 0x01) << 3);

	to[0] |= ((from[4] & 0x80) >> 3);
	to[1] |= ((from[4] & 0x40) >> 2);
	to[2] |= ((from[4] & 0x20) >> 1);
	to[3] |= ((from[4] & 0x10) >> 0);
	to[4] |= ((from[4] & 0x08) << 1);
	to[5] |= ((from[4] & 0x04) << 2);
	to[6] |= ((from[4] & 0x02) << 3);
	to[7] |= ((from[4] & 0x01) << 4);

	to[0] |= ((from[5] & 0x80) >> 2);
	to[1] |= ((from[5] & 0x40) >> 1);
	to[2] |= ((from[5] & 0x20) >> 0);
	to[3] |= ((from[5] & 0x10) << 1);
	to[4] |= ((from[5] & 0x08) << 2);
	to[5] |= ((from[5] & 0x04) << 3);
	to[6] |= ((from[5] & 0x02) << 4);
	to[7] |= ((from[5] & 0x01) << 5);

	to[0] |= ((from[6] & 0x80) >> 1);
	to[1] |= ((from[6] & 0x40) >> 0);
	to[2] |= ((from[6] & 0x20) << 1);
	to[3] |= ((from[6] & 0x10) << 2);
	to[4] |= ((from[6] & 0x08) << 3);
	to[5] |= ((from[6] & 0x04) << 4);
	to[6] |= ((from[6] & 0x02) << 5);
	to[7] |= ((from[6] & 0x01) << 6);

	to[0] |= ((from[7] & 0x80) >> 0);
	to[1] |= ((from[7] & 0x40) << 1);
	to[2] |= ((from[7] & 0x20) << 2);
	to[3] |= ((from[7] & 0x10) << 3);
	to[4] |= ((from[7] & 0x08) << 4);
	to[5] |= ((from[7] & 0x04) << 5);
	to[6] |= ((from[7] & 0x02) << 6);
	to[7] |= ((from[7] & 0x01) << 7);

}

unsigned short /*char*/ *GetFontData(FONTMETRICS *Font, unsigned short sjis_code){
	int		i;

	for (i=0; i<Font->Blocks; i++){
		if (sjis_code >= Font->Block[i].StartCode && sjis_code <= Font->Block[i].EndCode){
//			return (Font->Block[i].FontData + (sjis_code - Font->Block[i].StartCode) * Font->BytesPerCharWrk);
			return (Font->Block[i].FontData + (sjis_code - Font->Block[i].StartCode) * Font->WordsPerCharWrk);
		}
	}

	return NULL;
}


int InitFont(){

	printf("InitFont start\n");
	if (GetFontMetrics(FONT_FILE_ZEN_12_12, &FontZen_12_12) == 0){
		printf("GetFontMetrics error %s\n", FONT_FILE_ZEN_12_12);
		return 0;
	}
	if (GetFontMetrics(FONT_FILE_HAN_6_12, &FontHan_6_12) == 0){
		printf("GetFontMetrics error %s\n", FONT_FILE_HAN_6_12);
		return 0;
	}
	printf("InitFont end\n");
	return 1;
}


int MakeLcdData(unsigned short *line1, unsigned short *line2, int lcd_width, unsigned char *sjis_str, int str_length){
	int				i, j;
	int				x;
	FONTMETRICS		*pfzen;
	FONTMETRICS		*pfhan;
	unsigned short	sjis_char;
	unsigned short	*fontbyte;
	int				line;
	
	pfhan = &FontHan_6_12;
	pfzen = &FontZen_12_12;
	
	line = 1;
	x = 0;
	
//	printf("MakeLcdData=%s ", sjis_str);
	for (i=0; i<str_length; i++){
		//if (_ismbslead((unsigned char *)sjis_str, (unsigned char *)(sjis_str + i)) != 0){
		if (sjis_str[i] >= 0x80){
			sjis_char = ((sjis_str[i] << 8) | sjis_str[i+1]);
			fontbyte = GetFontData(pfzen, sjis_char);
			if (fontbyte != NULL){
				if ((x + pfzen->Width) > lcd_width){
					if (line == 1){
						line = 2;
						x = 0;
					} else break;
				}
				for (j=0; j<pfzen->Width; j++){
					if (line == 1){
						line1[x] = *(fontbyte + j);
					} else {
						line2[x] = *(fontbyte + j);
					}
					x++;
				}
			}
			i++;
		} else {
			fontbyte = GetFontData(pfhan, sjis_str[i]);
			if (fontbyte != NULL){
				if ((x + pfhan->Width) > lcd_width){
					if (line == 1){
						line = 2;
						x = 0;
					} else break;
				}
				for (j=0; j<pfhan->Width; j++){
					if (line == 1){
						line1[x] = *(fontbyte + j);
					} else {
						line2[x] = *(fontbyte + j);
					}
					x++;
				}
			}
		}
	}
}

