﻿
// Debian GNU/Linux on WSL
// gcc mp3_13_mkfl.c -lid3tag -lz
// Make file list

#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <iconv.h>
#include <id3tag.h>			// /usr/local/include
#include <time.h>

#include "mp3_13.h"

#define		SNDTAGINFO				"SNDTAG13.DAT"		// タグ情報出力先


// ディレクトリ構成解析作業用
typedef struct _DirList {
	char		displayname[DL_DISPLAYNAME_LEN + 1];	// DIRの表示名(スペース埋め)
	char		fulldir[DIRPATH_LEN + 1];		// DIRフルパス
	uint16_t	lenfulldir;						// 同バイト数
	uint16_t	recno_subdir_start;		// サブDIR開始行
	uint16_t	num_subdir;				// サブDIR数
	uint16_t	recno_parentdir;		// 親DIR行
	uint16_t	recno_grandparentdir;	// 親親DIR行
	uint16_t	recno_nextrandomplay;	// ランダム再生時、次のレコードNo
	uint16_t	num_sndfile;			// 音源ファイル数
	// 以下、インデックス作成用
	uint16_t	recno_dirlist;			// ディレクトリ構成ファイルでのレコード番号
	char		index[INDEX_LEN];		// 表示名のインデックス(頭文字)、SJIS
//	char	displaynameuc[DL_DISPLAYNAME_LEN + 1];		// DIRの表示名(ASCII文字を大文字化UPPER CASE INDEX生成に使う)
//	int		same_uc_flag;			// 同一のdisplaynameucが存在するとき 1
	
} DIRLIST;


// MP3ファイルリスト構造体
typedef struct _Mp3List {
	char	TitleName[ML_TITLENAME_LEN + 1];	// 表示名
	char	FileName[ML_FILENAME_LEN];
} MP3LIST13;


// インデックス
typedef struct _IndexList {
	char		displayname[INDEX_LEN + 1];	// インデックスの表示名(スペース埋め)
	//char		Index[INDEX_LEN];			// インデックスの表示名(スペース埋め)
	uint16_t	recno_subdir_start;			// サブDIR開始行
	uint16_t	num_subdir;					// サブDIR数
} INDEXLIST;


//
DIRLIST			DirList[MAX_DIRS];
uint16_t		ShffuleWork[MAX_DIRS];
uint16_t		DirListDetailRecNum;	// DIR構成ファイルの明細レコード総数
uint16_t		NumSndDir;
INDEXLIST		IndexList[MAX_INDEXES];
uint8_t			NumIndexes;
uint16_t		FirstRandomplay;

iconv_t icdUCS4toUTF8;
iconv_t icdUTF8toSJIS;
iconv_t icdUTF16LEtoSJIS;

int ReadDirectoryStructure();
int compare_DIRLIST(const void *a, const void *b);
int WriteDirStructureList();
uint16_t MakeSndFileList(char *inFolder);
int compare_MP3LIST13(const void *a, const void *b);
int iconvUCS4toUTF8(id3_ucs4_t *ucs4, char *utf8str, int bufsize);
int iconvUTF8toSJIS(char *utf8str, int convbytes, char *sjisstr, int sjismaxlen);
int iconvUTF16LEtoSJIS(char *ucs4str, int convbytes, char *sjisstr, int sjismaxlen);
int convUCS4toUTF16LE(id3_ucs4_t *ucs4, char *utf16lestr, int bufsize);
void convUCS4toASCII(id3_ucs4_t *ucs4, char *asciistr, int bufsize);
void getID3TagInfo(char *inFolder, char *mp3, char *title, int titlebufsize, int *titlebytes);
void getID3TagInfoTest(FILE *fout, char *inFolder, char *mp3);
int InitFont();
int MakeLcdData(unsigned short *line1, unsigned short *line2, int lcd_width, unsigned char *sjis_str, int str_length);
void CreateMenuData13();
int SetRandomOrder();
void CreateIndexList();
int compare_DIRLIST_index(const void *a, const void *b);
void WriteIndexList();
void CreateHistFile();
void CreateDigitImageDat();


int main(){
	icdUCS4toUTF8 = iconv_open("UTF-8", "UCS-4LE");
	icdUTF8toSJIS = iconv_open("Shift_JIS//TRANSLIT", "UTF-8");
	//icdUTF16LEtoSJIS = iconv_open("Shift_JIS//TRANSLIT", "UTF-16");

	InitFont();
	
	/*
	 * ディレクトリ構造を読み取る
	 */
	ReadDirectoryStructure();
	
	/*
	 *	ランダム再生の順序を決める
	 */
	SetRandomOrder();
	
	/*
	 *	ディレクトリ構成ファイルを書き出す
	 */
	WriteDirStructureList();
	
	/*
	 *	インデックスファイル
	 */
	CreateIndexList();
	WriteIndexList();

	// メインメニューイメージ
	CreateMenuData13();
	// 数字イメージファイル
	CreateDigitImageDat();
	
	// 履歴ファイル
	CreateHistFile();
	
	iconv_close(icdUCS4toUTF8);
	iconv_close(icdUTF8toSJIS);
	//iconv_close(icdUTF16LEtoSJIS);
}


/*
 *	ディレクトリ構造を読み取り、内部構造体にセットする。
 *	MP3ファイルの存在するディレクトリについて「音源ファイル(MP3)リスト」を作成していく。
 */
int ReadDirectoryStructure(){
	FILE	*fin;
	char	buff[1024];
	uint16_t	rowin;
	uint16_t	idxout;
	char	*c;
	char	*fullp;
	char	*parentp;
	char	*name;
	int		i, p;
	int		len;

	printf("ReadDirectoryStructure\n");
#if 0
	// linuxでは find /snd -maxdepth 10 -type d
	
	// DIRコマンド /AD=ディレクトリに関して、/ON=名前順、/S=サブディレクトリも、/B=ファイル名
	// << 同一ディレクトリ下のサブディレクトリは連続することが以下のプログラムの前提になっている。>>
	_stprintf(buff, _T("dir /AD /ON /S /N /B > %s%s"), TempDir, DIRLISTFILE);
	_tsystem(buff);
#endif

	//sprintf(buff, _T("%s"), TempDir, DIRLISTFILE);
	fin = fopen(DIRLISTFILE, "r");
	if (fin == NULL){
		printf("cannot open input file\n");
		return -1;
	}

	NumSndDir = 0;		// MP3ファイルを持つDIR数
	rowin = 0;
	idxout = 0;

	while (fgets(buff, 1024, fin) != NULL){
		printf("%s", buff);
		// 改行コードを取る
		c = strchr(buff, '\n');		// 0x0A
		if (c != NULL){
			*c = '\0';
		}
		fullp = buff;
		if (strlen(fullp) > DIRPATH_LEN) continue;

		// fullpはディレクトリのフルパスである。
		// これをパス無しサブディレクトリ名と親ディレクトリパスとに分離する
		c = strrchr(fullp, '/');
		if (c != NULL){
			strcpy(DirList[idxout].fulldir, fullp);
			DirList[idxout].lenfulldir = strlen(DirList[idxout].fulldir);
			*c = '\0';
			name = c + 1;
			if (name[0] == '_') continue;		// _ で始まるディレクトリは無視
			
			memset(DirList[idxout].displayname, ' ', DL_DISPLAYNAME_LEN);
			//		printf("iconvUTF8toSJIS start\n");
			iconvUTF8toSJIS(name, strlen(name), DirList[idxout].displayname, DL_DISPLAYNAME_LEN);
			//		printf("iconvUTF8toSJIS end\n");
			DirList[idxout].displayname[DL_DISPLAYNAME_LEN] = '\0';
			DirList[idxout].recno_subdir_start = 0;
			DirList[idxout].num_subdir = 0;
			DirList[idxout].recno_parentdir = 0;
			DirList[idxout].recno_grandparentdir = 0;
			DirList[idxout].recno_nextrandomplay = 0;
			DirList[idxout].num_sndfile = 0;
			parentp = fullp;

			// DirListから親DIRを探す
			for (i=0; i<idxout; i++){
				//printf("[%s][%s]\n", parentp, DirList[i].fulldir);
				if (strcmp(parentp, DirList[i].fulldir) == 0){
					DirList[idxout].recno_parentdir = i + 1;		// 親DIRの位置
					// +1しているのはqsort実行時にルートディレクトリとルートディレクトリ直下のディレクトリを区別するため。
					// ここでrecno_parentdirに設定する値は暫定値。後にDirListをソートしてから親DIRを再設定する
					break;
				}
			}
			if (i == idxout){
				// 親が見つからないのはidxout==0すなわちルートDIR
			}
			
			DirList[idxout].num_sndfile = MakeSndFileList(DirList[idxout].fulldir);		// 音源ファイルリストを生成

			rowin++;
			idxout++;
		}
	}
	fclose(fin);

	DirListDetailRecNum = idxout;
	NumSndDir = 0;
	
	// 親DIRが共通のディレクトリが連続するように並べ替える。
	qsort(DirList, DirListDetailRecNum, sizeof(DIRLIST), compare_DIRLIST);
	// 親DIRの再検索
	for (i=0; i<DirListDetailRecNum; i++){
		c = strrchr(DirList[i].fulldir, '/');
		if (c != NULL){
			// 親DIR
			memcpy(buff, DirList[i].fulldir, (int)(c - DirList[i].fulldir));
			buff[(int)(c - DirList[i].fulldir)] = '\0';
			// 親DIRを探す
			for (p=0; p<i; p++){
				if (strcmp(buff, DirList[p].fulldir) == 0){
					if (DirList[p].num_subdir == 0){
						DirList[p].recno_subdir_start = i;	// 親DIRの最初のサブDIR
					}
					DirList[p].num_subdir++;			// 親DIRのサブDIR数
					DirList[i].recno_parentdir = p;		// 親DIRの位置
					DirList[i].recno_grandparentdir = DirList[p].recno_parentdir;	// 親親DIRの位置
					break;
				}
			}
			
			// 
			if (DirList[i].num_sndfile > 0){
				// ランダム再生順決定テーブル
				ShffuleWork[NumSndDir] = i;
				NumSndDir++;
			}
		}
	}
	
	return 0;
}


// DIRLIST配列の並べ替え
int compare_DIRLIST(const void *a, const void *b){
	DIRLIST *p1 = (DIRLIST *) a;
	DIRLIST *p2 = (DIRLIST *) b;
	
	if (p1 != NULL && p2 != NULL){
		if (p1->recno_parentdir < p2->recno_parentdir) return -1;
		else if (p1->recno_parentdir > p2->recno_parentdir) return 1;
		//else return strcmp(p1->fulldir, p2->fulldir);
		else return memcmp(p1->displayname, p2->displayname, DL_DISPLAYNAME_LEN + 1);
	}
	return 0;
}

/*
 *	DIR構成ファイルを書き出す
 */
int WriteDirStructureList(){
	FILE			*fout;
	char			rec[RECLEN_DIRLIST];
	int				i;
	unsigned short	line1[LCD_WIDTH];
	unsigned short	line2[LCD_WIDTH];
	
	printf("WriteDirectoryStructureList\n");
	// DIR構成ファイル出力
	// ０レコード目はヘッダレコード
	fout = fopen(DIRSTRUCTUREFILE13, "w");
	memset(rec, '\0', RECLEN_DIRLIST);
	fwrite(&DirListDetailRecNum, 2, 1, fout);		// ヘッダ
	fwrite(&NumSndDir, 2, 1, fout);
	fwrite(rec, RECLEN_DIRLIST - 4, 1, fout);
	
	memset(line1, 0, LCD_WIDTH*2);
	memset(line2, 0, LCD_WIDTH*2);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	fwrite(line2, DL_LCDIMAGE, 1, fout);

	// DIR名イメージファイル
	// １レコード目からがDIRレコード（その最初のレコードはルート）
	for (i=0; i<DirListDetailRecNum; i++){
		//_tprintf(_T("%-12s %3d %04d %03d %04d %04d %04d %02d\n"), 
		//	DirList[i].dirname, i, DirList[i].recno_subdir_start, DirList[i].num_subdir, 
		//	DirList[i].recno_parentdir, DirList[i].recno_grandparentdir, DirList[i].recno_nextrandomplay, DirList[i].num_sndfile);

		if (DirList[i].recno_subdir_start != 0){	// サブディレクトリがある
			DirList[i].recno_subdir_start += 1;		// ヘッダレコードの分を加算
		}
		if (DirList[i].num_subdir > 0x00FF){
			printf("サブDIR数制限超過 %s %d\n", DirList[i].fulldir, DirList[i].num_subdir);
			DirList[i].num_subdir = 0x00FF;	// 最初の２５５だけ有効に。
			// 制限超過していてもランダム再生の対象にはなる。
			// DIR選択機能で選択ができない。
		}
		DirList[i].recno_parentdir += 1;	// ヘッダレコードの分を加算（ルートディレクトリは自分自身）
		DirList[i].recno_grandparentdir += 1;	// ヘッダレコードの分を加算（ルートディレクトリは自分自身）
		DirList[i].recno_nextrandomplay += 1;	// ヘッダレコードの分を加算
		DirList[i].recno_dirlist = i + 1;
		
		fwrite(&DirList[i].displayname, DL_DISPLAYNAME_LEN, 1, fout);
		fwrite(DirList[i].fulldir, DL_FULLDIR_LEN, 1, fout);
		fwrite(&DirList[i].lenfulldir, DL_LENFULLDIR_LEN, 1, fout);
		fwrite(&DirList[i].recno_subdir_start, DL_RECNO_SUBDIR_LEN, 1, fout);
		fwrite(&DirList[i].num_subdir, DL_NUM_SUBDIR_LEN, 1, fout);
		fwrite(&DirList[i].recno_parentdir, DL_RECNO_PARENTDIR_LEN, 1, fout);
		fwrite(&DirList[i].recno_grandparentdir, DL_RECNO_GRANDPARENTDIR_LEN, 1, fout);
		fwrite(&DirList[i].recno_nextrandomplay, DL_RECNO_NEXTRANDOMDIR_LEN, 1, fout);
		fwrite(&DirList[i].num_sndfile, DL_NUM_SNDFILE_LEN, 1, fout);
		fwrite(rec, DL_FILLER_13, 1, fout);
		
		memset(line1, 0, LCD_WIDTH*2);
		memset(line2, 0, LCD_WIDTH*2);
		MakeLcdData(line1, line2, LCD_WIDTH, DirList[i].displayname, DL_DISPLAYNAME_LEN);
		fwrite(line1, DL_LCDIMAGE, 1, fout);
		fwrite(line2, DL_LCDIMAGE, 1, fout);
		
	}
	fclose(fout);

	return 0;
}


/*
 *	音源ファイル(MP3)リスト作成
 */
uint16_t MakeSndFileList(char *inFolder){
	DIR			*dirp;
	struct dirent *dp;
	char		*ext;
	FILE		*fout;
	char		outfile[DIRPATH_LEN + MP3FILE_LEN + 1];
	char		title[256];
	int			titlebytes;
	unsigned short	line1[LCD_WIDTH];
	unsigned short	line2[LCD_WIDTH];
	uint16_t	num = 0;
	MP3LIST13	rec04[0x100];
	char		rec[RECLEN_MP3LIST];
	int			i, j, k, len;
	char		TitleNamePrefix[ML_TITLENAME_LEN + 1];
	int			PrefixLen = 0;
	//int			CommonPrefix = 0;
	
	//printf("MakeSndFileList %s\n", inFolder);
	dirp = opendir(inFolder);
	if (dirp == NULL) return 0;
	
	// タグ情報取得テストの出力ファイル
	sprintf(outfile, "%s/%s", inFolder, SNDTAGINFO);
	fout = fopen(outfile, "w");
	
	memset(rec, 0, RECLEN_MP3LIST);
	
	while ((dp = readdir(dirp)) != NULL){
		//printf("MakeSndFileList=%s\n", dp->d_name);
		if (dp->d_type == DT_REG){
			ext = strrchr(dp->d_name, '.');		// 拡張子
			if (ext != NULL){
				if (strcmp(ext, ".mp3") == 0 || strcmp(ext, ".MP3") == 0){
					if (strlen(dp->d_name) > MP3FILE_LEN){
						continue;
					}
					if (num == 0x00FF){
						printf("MP3ファイル数制限超過 %s\n", inFolder);
						break;
					}
					
					// タグ情報取得テスト
					//	printf("getID3TagInfoTest start\n");
					getID3TagInfoTest(fout, inFolder, dp->d_name);
					//	printf("getID3TagInfoTest end\n");
					
					memset(&rec04[num], 0, sizeof(MP3LIST13));
					strcpy(rec04[num].FileName, dp->d_name);
					memset(rec04[num].TitleName, ' ', ML_TITLENAME_LEN);
					rec04[num].TitleName[ML_TITLENAME_LEN] = '\0';
					
					num++;
				}
			}
		}
	}
	closedir(dirp);
	fclose(fout);
	
	//printf("num=%d\n", num);
	if (num > 0){
		// ファイル名の順にソート
		qsort(rec04, num, sizeof(MP3LIST13), compare_MP3LIST13);
		// タグ情報からタイトル取得
		for (i=0; i<num; i++){
			sprintf(title, "%02d ", i + 1);
			len = strlen(title);
			getID3TagInfo(inFolder, rec04[i].FileName, title + len, 255 - len, &titlebytes);
			if (titlebytes > 0){
				title[titlebytes + len] = '\0';
				//printf("  UTF8 : %s\n", title);
				iconvUTF8toSJIS(title, titlebytes + len, rec04[i].TitleName, ML_TITLENAME_LEN);
			} else {
				ext = strrchr(rec04[i].FileName, '.');		// 拡張子
				iconvUTF8toSJIS(rec04[i].FileName, (int)(ext - rec04[i].FileName), rec04[i].TitleName, ML_TITLENAME_LEN);
			}
			// 全ファイル名の共通プレフィックス取得
			if (i == 1){
				for (j=0; j<ML_TITLENAME_LEN; j++){
					if (rec04[0].TitleName[j] == '\0' || rec04[1].TitleName[j] == '\0' ||
						rec04[0].TitleName[j] == '.' || rec04[1].TitleName[j] == '.'){
						break;
					} else if (rec04[0].TitleName[j] == rec04[1].TitleName[j]){
						TitleNamePrefix[j] = rec04[0].TitleName[j];
					} else {
						TitleNamePrefix[j] = '\0';
						PrefixLen = j;
						break;
					}
				}
			} else if (i > 1 && PrefixLen > 0){
				for (j=0; j<PrefixLen; j++){
					if (rec04[i].TitleName[j] == '\0' || rec04[i].TitleName[j] == '.'){
						PrefixLen = 0;
						break;
					} else if (rec04[i].TitleName[j] == TitleNamePrefix[j]){
					} else {
						TitleNamePrefix[j] = '\0';
						PrefixLen = j;
						break;
					}
				}
			}
		}
		// 共通プレフィックスがあるときは、それを除いた部分をタイトル名とする
		if (PrefixLen > 0){
			char	TitleNameNoPrefix[ML_TITLENAME_LEN + 1];	// 表示名
			for (i=0; i<num; i++){
				memset(TitleNameNoPrefix, ' ', ML_TITLENAME_LEN);
				TitleNameNoPrefix[ML_TITLENAME_LEN] = '\0';
				memcpy(TitleNameNoPrefix, rec04[i].TitleName + PrefixLen, ML_TITLENAME_LEN - PrefixLen);
				memcpy(rec04[i].TitleName, TitleNameNoPrefix, ML_TITLENAME_LEN);
			}
		}
	}
	
	// 音源ファイルリストファイル出力
	sprintf(outfile, "%s/%s", inFolder, SNDFILELISTFILE);
	fout = fopen(outfile, "w");
	for (i=0; i<num; i++){
		fwrite(&rec04[i].TitleName, ML_TITLENAME_LEN, 1, fout);
		fwrite(&rec04[i].FileName, ML_FILENAME_LEN, 1, fout);
		fwrite(rec, ML_FILLER_13, 1, fout);
		
		memset(line1, 0, LCD_WIDTH*2);
		memset(line2, 0, LCD_WIDTH*2);
		MakeLcdData(line1, line2, LCD_WIDTH, rec04[i].TitleName, ML_TITLENAME_LEN);
		fwrite(line1, DL_LCDIMAGE, 1, fout);
		fwrite(line2, DL_LCDIMAGE, 1, fout);
	}
	fclose(fout);
	
	return num;
}


// MP3LIST13配列の並べ替え
int compare_MP3LIST13(const void *a, const void *b){
	MP3LIST13 *p1 = (MP3LIST13 *) a;
	MP3LIST13 *p2 = (MP3LIST13 *) b;
	
	if (p1 != NULL && p2 != NULL){
		return strcmp(p1->FileName, p2->FileName);
	}
	return 0;
}


/*
 *	mp3ファイルのタグ情報から必要なものを取得
 *		title : 曲のタイトルをUTF8で取得する。末尾にNULL文字をセットしない。
 *		titlebufsize : titleバッファのサイズ
 *		*titlebytes : タイトルのバイト数。タイトルを取得できないときはマイナス。
 *		
 *	https://www.mars.org/mailman/public/mad-dev/2002-January/000439.html
 */
void getID3TagInfo(char *inFolder, char *mp3, char *title, int titlebufsize, int *titlebytes){
	char		mp3file[DIRPATH_LEN + MP3FILE_LEN + 1];
	struct id3_file *id3_file;
	struct id3_tag *id3_tag;
	struct id3_frame *id3_frame;
	int			i, j, k;
	
	*titlebytes = -1;
	
	sprintf(mp3file, "%s/%s", inFolder, mp3);
	id3_file = id3_file_open(mp3file, ID3_FILE_MODE_READONLY);
	id3_tag = id3_file_tag(id3_file);
	for (i=0; i<id3_tag->nframes; i++){
		id3_frame = id3_tag->frames[i];
		if (id3_frame != NULL){
			if (strcmp(id3_frame->id, ID3_FRAME_TITLE) == 0){
				for (j=0; j<id3_frame->nfields; j++){
					union id3_field		id3_field;
					id3_field = id3_frame->fields[j];
					
					if (id3_field.type == ID3_FIELD_TYPE_STRINGLIST){
						for (k=0; k<id3_field_getnstrings(&id3_field); k++){
							id3_ucs4_t *ucs4_str = id3_field_getstrings(&id3_field, k);
							*titlebytes = iconvUCS4toUTF8(ucs4_str, title, titlebufsize);
						}
					}
				}
			}
		}
	}
	id3_file_close(id3_file);
}

/*
 *	mp3ファイルのID3タグ情報を取得するテストプログラム
 *	(/usr/local/include/id3tag.h参照)
 *	https://www.mars.org/mailman/public/mad-dev/2002-January/000439.html
 */
void getID3TagInfoTest(FILE *fout, char *inFolder, char *mp3){
	char		mp3file[DIRPATH_LEN + MP3FILE_LEN + 1];
	struct id3_file *id3_file;
	struct id3_tag *id3_tag;
	struct id3_frame *id3_frame;
	char		utfbuf[512];
	char		sjisbuf[512];
	int			i, j, k;
	
	sprintf(mp3file, "%s/%s", inFolder, mp3);
	id3_file = id3_file_open(mp3file, ID3_FILE_MODE_READONLY);
	id3_tag = id3_file_tag(id3_file);
	//printf("id3_tag nframes=%d\n", id3_tag->nframes);
	for (i=0; i<id3_tag->nframes; i++){
		//signed long encoding = 0;
		
		id3_frame = id3_tag->frames[i];
		if (id3_frame != NULL){
			//printf("%d ID=%s desc=%s nfields=%d\n", i, id3_frame->id, id3_frame->description, id3_frame->nfields);
			fprintf(fout, "Frame Number = %d ID=%s\n", i, id3_frame->id);
			for (j=0; j<id3_frame->nfields; j++){
				union id3_field		id3_field;
				id3_field = id3_frame->fields[j];
				//printf(" field=%d id3_field_type=%d \n", j, id3_field.type);
				switch (id3_field.type){
				case ID3_FIELD_TYPE_TEXTENCODING: 
					//printf("ID3_FIELD_TYPE_TEXTENCODING "); 
					//printf("value=%ld ", id3_field.number.value);
					//encoding = id3_field.number.value;	// encodingを覚えておく必要はない
					break;
				case ID3_FIELD_TYPE_LATIN1: 
					//printf("ID3_FIELD_TYPE_LATIN1 "); 
					fprintf(fout, " ID3_FIELD_TYPE_LATIN1 : %s\n", id3_field.latin1.ptr);
					break;
				case ID3_FIELD_TYPE_LATIN1FULL: 
					//printf("ID3_FIELD_TYPE_LATIN1FULL "); 
					fprintf(fout, " ID3_FIELD_TYPE_LATIN1FULL : %s\n", id3_field.latin1.ptr);
					break;
				case ID3_FIELD_TYPE_LATIN1LIST: 
					//printf("ID3_FIELD_TYPE_LATIN1LIST "); 
					for (k=0; k<id3_field.latin1list.nstrings; k++){
						fprintf(fout, " ID3_FIELD_TYPE_LATIN1LIST : %s\n", id3_field.latin1list.strings[k]);
					}
					break;
				case ID3_FIELD_TYPE_STRING: 
					//printf("ID3_FIELD_TYPE_STRING "); 
					//fprintf(fout, " ID3_FIELD_TYPE_STRING : %s\n", id3_field.string.ptr);
					{
						id3_ucs4_t *ucs4_str = id3_field_getstring(&id3_field);
						//if (encoding == ID3_FIELD_TEXTENCODING_UTF_16){
							//int bytes = convUCS4toUTF16LE(ucs4_str, utfbuf, 128);
							//iconvUTF16LEtoSJIS(utfbuf, bytes, sjisbuf, 128);
							int bytes = iconvUCS4toUTF8(ucs4_str, utfbuf, 512);
							bytes = iconvUTF8toSJIS(utfbuf, bytes, sjisbuf, 512);
							sjisbuf[bytes] = '\0';
							fprintf(fout, " ID3_FIELD_TYPE_STRING : %s\n", sjisbuf);
						//} else if (encoding == ID3_FIELD_TEXTENCODING_ISO_8859_1){
						//	convUCS4toASCII(ucs4_str, sjisbuf, 128);
						//	fprintf(fout, " ID3_FIELD_TYPE_STRING : %s\n", sjisbuf);
						//}
					}
					break;
				case ID3_FIELD_TYPE_STRINGFULL: 
					//printf("ID3_FIELD_TYPE_STRINGFULL "); 
					//fprintf(fout, " ID3_FIELD_TYPE_STRINGFULL : %s\n", id3_field.string.ptr);
					{
						id3_ucs4_t *ucs4_str = id3_field_getfullstring(&id3_field);
						//if (encoding == ID3_FIELD_TEXTENCODING_UTF_16){
							//int bytes = convUCS4toUTF16LE(ucs4_str, utfbuf, 128);
							//iconvUTF16LEtoSJIS(utfbuf, bytes, sjisbuf, 128);
							int bytes = iconvUCS4toUTF8(ucs4_str, utfbuf, 512);
							bytes = iconvUTF8toSJIS(utfbuf, bytes, sjisbuf, 512);
							sjisbuf[bytes] = '\0';
							fprintf(fout, " ID3_FIELD_TYPE_STRINGFULL : %s\n", sjisbuf);
						//} else if (encoding == ID3_FIELD_TEXTENCODING_ISO_8859_1){
						//	convUCS4toASCII(ucs4_str, sjisbuf, 128);
						//	fprintf(fout, " ID3_FIELD_TYPE_STRINGFULL : %s\n", sjisbuf);
						//}
					}
					break;
				case ID3_FIELD_TYPE_STRINGLIST: 
					//printf("ID3_FIELD_TYPE_STRINGLIST "); 
					//printf("nstrings=%d ", id3_field.stringlist.nstrings);
					for (k=0; k<id3_field_getnstrings(&id3_field) /*id3_field.stringlist.nstrings*/; k++){
						id3_ucs4_t *ucs4_str = id3_field_getstrings(&id3_field, k);
						//if (encoding == ID3_FIELD_TEXTENCODING_UTF_16){
							//int bytes = convUCS4toUTF16LE(ucs4_str, utfbuf, 128);
							//iconvUTF16LEtoSJIS(utfbuf, bytes, sjisbuf, 128);
							int bytes = iconvUCS4toUTF8(ucs4_str, utfbuf, 512);
							bytes = iconvUTF8toSJIS(utfbuf, bytes, sjisbuf, 512);
							sjisbuf[bytes] = '\0';
							fprintf(fout, " ID3_FIELD_TYPE_STRINGLIST : %s\n", sjisbuf);
						//} else if (encoding == ID3_FIELD_TEXTENCODING_ISO_8859_1){
						//	printf("  convUCS4toASCII start\n");
						//	convUCS4toASCII(ucs4_str, sjisbuf, 128);
						//	printf("  convUCS4toASCII end\n");
						//	fprintf(fout, " ID3_FIELD_TYPE_STRINGLIST : %s\n", sjisbuf);
						//}
					}
					break;
				case ID3_FIELD_TYPE_LANGUAGE: 
					//printf("ID3_FIELD_TYPE_LANGUAGE "); 
					/*
					{
						id3_ucs4_t *ucs4_str = id3_field.string.ptr;
						if (encoding == ID3_FIELD_TEXTENCODING_UTF_16){
							convUCS4toUTF16LE(ucs4_str, utfbuf, 128);
							iconvUTF16LEtoSJIS(utfbuf, 128, sjisbuf, 128);
							fprintf(fout, " ID3_FIELD_TYPE_LANGUAGE : %s\n", sjisbuf);
						} else if (encoding == ID3_FIELD_TEXTENCODING_ISO_8859_1){
							convUCS4toASCII(ucs4_str, sjisbuf, 128);
							fprintf(fout, " ID3_FIELD_TYPE_LANGUAGE : %s\n", sjisbuf);
						}
					}
					*/
					break;
				case ID3_FIELD_TYPE_FRAMEID: 
					//printf("ID3_FIELD_TYPE_FRAMEID "); 
					break;
				case ID3_FIELD_TYPE_DATE: 
					//printf("ID3_FIELD_TYPE_DATE "); 
					break;
				case ID3_FIELD_TYPE_INT8: 
					//printf("ID3_FIELD_TYPE_INT8 "); 
					break;
				case ID3_FIELD_TYPE_INT16: 
					//printf("ID3_FIELD_TYPE_INT16 "); 
					break;
				case ID3_FIELD_TYPE_INT24: 
					//printf("ID3_FIELD_TYPE_INT24 "); 
					break;
				case ID3_FIELD_TYPE_INT32: 
					//printf("ID3_FIELD_TYPE_INT32 "); 
					break;
				case ID3_FIELD_TYPE_INT32PLUS: 
					//printf("ID3_FIELD_TYPE_INT32PLUS "); 
					break;
				case ID3_FIELD_TYPE_BINARYDATA: 
					//printf("ID3_FIELD_TYPE_BINARYDATA "); 
					break;
				}
				//printf(" \n");
			}
		}
	}
	id3_file_close(id3_file);
}


/*
 *	ランダム再生のオーダーを決める
 */
int SetRandomOrder(){
	int				shuffle_int;
	int				shuffle_cnt;
	int				i, j;
	uint16_t		dirno_last;
	unsigned int	seed;
	int 			idx1, idx2;
	uint16_t		d;

	printf("random order start\n");
	// ランダム再生順決定
	if (NumSndDir > 50){
		shuffle_cnt = NumSndDir*1000;
		shuffle_int = NumSndDir / 10;
	} else {
		shuffle_cnt = 0;
		shuffle_int = 0;
	}
	seed = (unsigned)time(NULL);
	printf("seed=%d\n", seed);
	srand(seed);	// seedはこれでよいか
	for (i=0; i<shuffle_cnt; i++){
		idx1 = (rand() % NumSndDir);
		while (1){
			idx2 = (rand() % NumSndDir);
			// DirList[]内でなるベく離れた者同士を交換する
			if (abs(ShffuleWork[idx2] - ShffuleWork[idx1]) >= shuffle_int){
				break;
			}
		}
		d = ShffuleWork[idx1];
		ShffuleWork[idx1] = ShffuleWork[idx2];
		ShffuleWork[idx2] = d;
	}
	if (NumSndDir > 50){
		for (i=1; i<(NumSndDir - 1); i++){
			// 親ディレクトリが共通するサブディレクトリが連続するのを避ける
			if (DirList[ShffuleWork[i]].recno_parentdir == DirList[ShffuleWork[i - 1]].recno_parentdir){	// 親ディレクトリが共通
				idx1 = i;
				//while (1){			// ディレクトリ構成によっては無限ループになる
				for (j=0; j<20; j++){	// 適当なところであきらめる
					idx2 = (rand() % NumSndDir);
					if (idx2 > 0 && idx2 < (NumSndDir - 1)){
						if (DirList[ShffuleWork[idx2]].recno_parentdir != DirList[ShffuleWork[i - 1]].recno_parentdir
							&& DirList[ShffuleWork[idx2]].recno_parentdir != DirList[ShffuleWork[i + 1]].recno_parentdir
							&& DirList[ShffuleWork[idx1]].recno_parentdir != DirList[ShffuleWork[idx2 - 1]].recno_parentdir
							&& DirList[ShffuleWork[idx1]].recno_parentdir != DirList[ShffuleWork[idx2 + 1]].recno_parentdir){
							
							d = ShffuleWork[idx1];
							ShffuleWork[idx1] = ShffuleWork[idx2];
							ShffuleWork[idx2] = d;
							//printf("%d, %d %d %d, %d %d %d\n", j, 
							//	DirList[ShffuleWork[idx2 - 1]].recno_parentdir, DirList[ShffuleWork[idx2]].recno_parentdir, DirList[ShffuleWork[idx2 + 1]].recno_parentdir,
							//	DirList[ShffuleWork[idx1 - 1]].recno_parentdir, DirList[ShffuleWork[idx1]].recno_parentdir, DirList[ShffuleWork[idx1 + 1]].recno_parentdir);
							break;
						}
					}
				}
			}
		}
	}
	//for (i=0; i<NumSndDir; i++) printf("%d ", ShffuleWork[i]);
	//printf("\n");
	dirno_last = 0;
	for (i=0; i<NumSndDir; i++){
		DirList[dirno_last].recno_nextrandomplay = ShffuleWork[i];
		//printf("shaffle=%d=%s\n", i, DirList[ShffuleWork[i]].fulldir);
		dirno_last = ShffuleWork[i];
	}
	DirList[dirno_last].recno_nextrandomplay = DirList[0].recno_nextrandomplay;		// 循環させる

	FirstRandomplay = DirList[0].recno_nextrandomplay;
	
	return 0;
}


/*
 *	インデックスファイルを作成する。
 *	インデックスファイルの構造はDIR構成ファイルに類似。
 *	ルートディレクトリ - インデックス - 実ディレクトリ　の３階層からなるDIR構成ファイルと考えられる。
 */
void CreateIndexList(){
	int				i, j;
	unsigned char	index[INDEX_LEN];
	
	printf("CreateIndexList start\n");
	// MP3ファイルを持つ各ディレクトリのインデックスを抽出
	for (i=0; i<DirListDetailRecNum; i++){
		if (DirList[i].num_sndfile > 0){
			memset(index, ' ', INDEX_LEN);
			index[0] = DirList[i].displayname[0];
			if (index[0] < 0x80){		// 1バイト文字
				if (index[0] >= 'a' && index[0] <= 'z'){
					index[0] += ('A' - 'a');	// 大文字で
				}
			} else {
				index[1] = DirList[i].displayname[1];
			}
			memcpy(DirList[i].index, index, INDEX_LEN);
		}
	}
	
	// INDEXで並べ替え
	qsort(DirList, DirListDetailRecNum, sizeof(DIRLIST), compare_DIRLIST_index);
	//for (i=0; i<DirListDetailRecNum; i++){
	//	if (DirList[i].num_sndfile > 0){
	//		printf("%s %s\n", DirList[i].index, DirList[i].fulldir);
	//	}
	//}

#if 0
	// displaynameucの重複チェック(たとえばdisk1のようなディレクトリ名は複数ありえる)
	memset(displayname, 0, DL_DISPLAYNAME_LEN + 1);
	for (i=0; i<DirListDetailRecNum; i++){
		if (DirList[i].num_sndfile > 0){
			if (memcmp(displayname, DirList[i].displayname) == 0){
				DirList[i].same_uc_flag = 1;
				DirList[i - 1].same_uc_flag = 1;
			} else {
				DirList[i].same_uc_flag = 0;
			}
			memcpy(displayname, DirList[i].displayname);
		}
	}
	
	// 重複のあるdisplaynameについて、親ディレクトリ名|displayname を表示名とする。
	for (i=0; i<DirListDetailRecNum; i++){
		if (DirListWork[i].num_sndfile > 0){
			if (DirListWork[i].same_uc_flag == 1){
				int len = _tcslen(DirListWork[i].displayname);
				if (MAXLEN_DISPLAYNAME > (len+1)){
					DirListWork[i].displayname[len] = '|';
					memcpy(&DirListWork[i].displayname[len + 1], DirListWork[i].displaynameparent, MAXLEN_DISPLAYNAME - (len + 1));
				}
				printf("dupname=%s\n", DirListWork[i].displayname);
			}
		}
	}
#endif

	// インデックスごとに集計->IndexList作成
	memcpy(IndexList[0].displayname, DirList[0].index, INDEX_LEN);
	IndexList[0].num_subdir = 0;
	IndexList[0].recno_subdir_start = 0;	// DirListのindex
	NumIndexes = 1;
	
	for (i=1; i<DirListDetailRecNum; i++){
		if (DirList[i].num_sndfile > 0){
			if (memcmp(DirList[i].index, IndexList[NumIndexes - 1].displayname, INDEX_LEN) != 0){
				memcpy(IndexList[NumIndexes].displayname, DirList[i].index, INDEX_LEN);
				IndexList[NumIndexes].num_subdir = 0;
				IndexList[NumIndexes].recno_subdir_start = 0;
				NumIndexes++;
			}
		}
	}
	
	printf("IndexList %d *************************************************************\n", NumIndexes);
	//for (i=0; i<NumIndexes; i++){
	//	printf("%d %s\n", i, IndexList[i].displayname);
	//}

	for (i=0; i<DirListDetailRecNum; i++){
		if (DirList[i].num_sndfile > 0){
			for (j=0; j<NumIndexes; j++){
				if (memcmp(DirList[i].index, IndexList[j].displayname, INDEX_LEN) == 0){
					//printf(" %s %d\n", DirList[i].fulldir, j);
					DirList[i].recno_parentdir = j;
					DirList[i].recno_grandparentdir = 0;
					
					if (IndexList[j].num_subdir == 0){
						IndexList[j].recno_subdir_start = i;
					}
					IndexList[j].num_subdir++;
					break;
				}
			}
		}
	}
	printf("DirList for Index complete ******************************************\n");
//	for (i=0; i<DirListDetailRecNum; i++){
//		if (DirList[i].num_sndfile > 0){
//			printf("%s [%s] %d %d \n", DirList[i].index, DirList[i].fulldir, DirList[i].recno_parentdir, DirList[i].recno_grandparentdir);
//		}
//	}
}


/*
 *	インデックスファイルを出力する
 */
void WriteIndexList(){
	FILE			*fout;
	char			rec[RECLEN_DIRLIST];
	unsigned short	line1[LCD_WIDTH];
	unsigned short	line2[LCD_WIDTH];
	unsigned short	details;
	int				i, j;
	uint16_t		lenfulldir;
	uint16_t		recno_subdir_start;
	uint8_t			num_subdir;
	uint16_t		recno_parentdir;		// 親DIR行
	uint16_t		recno_grandparentdir;	// 親親DIR行
	uint16_t		recno_nextrandomplay;	// ランダム再生時、次のレコードNo
	uint8_t			num_sndfile;			// 音源ファイル数
	uint16_t		recno_dirlist;			// ディレクトリ構成ファイルでのレコード番号
	
	fout = fopen(INDEXFILE, "w");
	
	memset(rec, '\0', RECLEN_DIRLIST);
	// ０レコード目はヘッダレコード
	details = NumIndexes + NumSndDir;
	fwrite(&details, 2, 1, fout);
	fwrite(&NumSndDir, 2, 1, fout);
	for (i=0; i<(RECLEN_DIRLIST_FULL - 4); i++) fputc(0, fout);
	
	// １レコード目はルートディレクトリ
	lenfulldir = 0;
	recno_subdir_start = 2;
	num_subdir = NumIndexes;
	recno_parentdir = 0;
	recno_grandparentdir = 0;
	recno_nextrandomplay = 0;
	num_sndfile = 0;
	recno_dirlist = 0;
	
	for (i=0; i<DL_DISPLAYNAME_LEN; i++) fputc(' ', fout);
	for (i=0; i<DL_FULLDIR_LEN; i++) fputc(0, fout);
	fwrite(&lenfulldir, 2, 1, fout);
	fwrite(&recno_subdir_start, 2, 1, fout);
	fwrite(&num_subdir, 1, 1, fout);
	fwrite(&recno_parentdir, 2, 1, fout);
	fwrite(&recno_grandparentdir, 2, 1, fout);
	for (i=0; i<DL_RECNO_NEXTRANDOMDIR_LEN; i++) fputc(0, fout);
	fwrite(&num_sndfile, 1, 1, fout);
	fwrite(&recno_dirlist, 2, 1, fout);
	for (i=0; i<IL_FILLER_13; i++) fputc(0, fout);
	for (i=0; i<LCD_WIDTH*2; i++) fputc(0, fout);
	for (i=0; i<LCD_WIDTH*2; i++) fputc(0, fout);
	
	// インデックス
	memset(rec, ' ', DL_DISPLAYNAME_LEN);
	memset(rec + DL_DISPLAYNAME_LEN, 0, RECLEN_DIRLIST - DL_DISPLAYNAME_LEN);
	recno_subdir_start += NumIndexes;
	for (i=0; i<NumIndexes; i++){
		fwrite(&IndexList[i].displayname, INDEX_LEN, 1, fout);
		for (j=0; j<DL_DISPLAYNAME_LEN - INDEX_LEN; j++) fputc(' ', fout);
		
		for (j=0; j<DL_FULLDIR_LEN; j++) fputc(0, fout);
		
		lenfulldir = 0;
		num_subdir = IndexList[i].num_subdir;
		recno_parentdir = 1;		// ルートディレクトリ
		recno_grandparentdir = 1;	// ルートディレクトリ
		recno_nextrandomplay = 0;
		num_sndfile = 0;
		recno_dirlist = 0;
		
		fwrite(&lenfulldir, 2, 1, fout);
		fwrite(&recno_subdir_start, 2, 1, fout);
		fwrite(&num_subdir, 1, 1, fout);
		fwrite(&recno_parentdir, 2, 1, fout);
		fwrite(&recno_grandparentdir, 2, 1, fout);
		for (j=0; j<DL_RECNO_NEXTRANDOMDIR_LEN; j++) fputc(0, fout);
		fwrite(&num_sndfile, 1, 1, fout);
		fwrite(&recno_dirlist, 2, 1, fout);
		for (j=0; j<IL_FILLER_13; j++) fputc(0, fout);
		
		memset(line1, 0, LCD_WIDTH*2);
		memset(line2, 0, LCD_WIDTH*2);
		MakeLcdData(line1, line2, LCD_WIDTH, IndexList[i].displayname, INDEX_LEN);
		fwrite(line1, DL_LCDIMAGE, 1, fout);
		fwrite(line2, DL_LCDIMAGE, 1, fout);
		
		recno_subdir_start += IndexList[i].num_subdir;
	}
	
	// 実ディレクトリ
	for (i=0; i<DirListDetailRecNum; i++){
		if (DirList[i].num_sndfile > 0){
			DirList[i].recno_parentdir += 2;	// ヘッダレコード＋ルートレコードの分を加算
			DirList[i].recno_grandparentdir = 1;	// 実ディレクトリの親親ディレクトリはすべてルートディレクトリ
			
			fwrite(&DirList[i].displayname, DL_DISPLAYNAME_LEN, 1, fout);
			fwrite(DirList[i].fulldir, DL_FULLDIR_LEN, 1, fout);
			fwrite(&DirList[i].lenfulldir, DL_LENFULLDIR_LEN, 1, fout);
			fwrite(&DirList[i].recno_subdir_start, DL_RECNO_SUBDIR_LEN, 1, fout);
			fwrite(&DirList[i].num_subdir, DL_NUM_SUBDIR_LEN, 1, fout);
			fwrite(&DirList[i].recno_parentdir, DL_RECNO_PARENTDIR_LEN, 1, fout);
			fwrite(&DirList[i].recno_grandparentdir, DL_RECNO_GRANDPARENTDIR_LEN, 1, fout);
			fwrite(&recno_nextrandomplay, DL_RECNO_NEXTRANDOMDIR_LEN, 1, fout);
			fwrite(&DirList[i].num_sndfile, DL_NUM_SNDFILE_LEN, 1, fout);
			fwrite(&DirList[i].recno_dirlist, IL_DLRECNO_LEN, 1, fout);
			for (j=0; j<IL_FILLER_13; j++) fputc(0, fout);
			
			memset(line1, 0, LCD_WIDTH*2);
			memset(line2, 0, LCD_WIDTH*2);
			MakeLcdData(line1, line2, LCD_WIDTH, DirList[i].displayname, DL_DISPLAYNAME_LEN);
			fwrite(line1, DL_LCDIMAGE, 1, fout);
			fwrite(line2, DL_LCDIMAGE, 1, fout);
		}
	}
	
	fclose(fout);
}


// DIRLIST配列の並べ替え
int compare_DIRLIST_index(const void *a, const void *b){
	DIRLIST *p1 = (DIRLIST *) a;
	DIRLIST *p2 = (DIRLIST *) b;
	
	if (p1 != NULL && p2 != NULL){
		if (p1->num_sndfile > 0 && p2->num_sndfile == 0) return -1;
		else if (p1->num_sndfile == 0 && p2->num_sndfile > 0) return 1;
		else {
			int cmp = memcmp(p1->index, p2->index, INDEX_LEN);
			if (cmp == 0){
				return memcmp(p1->displayname, p2->displayname, DL_DISPLAYNAME_LEN + 1);
			} else {
				return cmp;
			}
		} 
	}
	return 0;
}

/*
 *	UCS4文字列をUTF8文字列に変換。
 *		ucs4 : UCS4文字列。終端NULLを前提。
 *		utf8str : 変換されたUTF8文字列。終端にNULLをセットしない。
 *		戻り値 : utf8strにセットされた文字列のバイト数。エラー時はマイナス。
 */
int iconvUCS4toUTF8(id3_ucs4_t *ucs4, char *utf8str, int bufsize){
	char *p_src, *p_dst;
	size_t n_src, n_dst;
	int	rc;
	int bytes = 0;
	
	p_dst = utf8str;
	n_dst = bufsize;
	
	while (*ucs4 != 0){
		p_src = (char *)ucs4;
		n_src = 4;
		if ((rc = iconv(icdUCS4toUTF8, &p_src, &n_src, &p_dst, &n_dst)) == -1){
			return -1;
		}
		bytes += rc;
		ucs4++;
	}
	
	return (bufsize - n_dst);
}

/*
 *	UTF8文字列をSJIS文字列に変換
 *		utf8str : 変換元文字列
 *		convbytes : 変換対象バイト数
 *		sjisstr : 返還後文字列用変数（末尾にNULLを付けない）
 *		sjismaxlen : 最大変換後バイト数
 */
int iconvUTF8toSJIS(char *utf8str, int convbytes, char *sjisstr, int sjismaxlen){
	char *p_src, *p_dst;
	size_t n_src, n_dst;
	int	rc;
	
	p_src = utf8str;
	p_dst = sjisstr;
	n_src = convbytes;
    n_dst = sjismaxlen;
	while(0 < n_src && 0 < n_dst){
		if ((rc = iconv(icdUTF8toSJIS, &p_src, &n_src, &p_dst, &n_dst)) == -1){
			return -1;
		}
	}
	return (sjismaxlen - n_dst);
}


/*
 *	UTF16LE文字列をSJIS文字列に変換
 *		utf8str : 変換元文字列
 *		convbytes : 変換対象バイト数
 *		sjisstr : 返還後文字列用変数
 *		sjismaxlen : 最大変換後バイト数
 */
int iconvUTF16LEtoSJIS(char *utf16str, int convbytes, char *sjisstr, int sjismaxlen){
	char *p_src, *p_dst;
	size_t n_src, n_dst;
	int	rc;
	
	p_src = utf16str;
	p_dst = sjisstr;
	n_src = convbytes;
    n_dst = sjismaxlen;
	while(0 < n_src && 0 < n_dst){
		//printf("   %ld %ld\n", n_src, n_dst);
		if ((rc = iconv(icdUTF16LEtoSJIS, &p_src, &n_src, &p_dst, &n_dst)) == -1){
			break;
		}
	}
	*p_dst = '\0';
	
	return rc;
}


// ucs4文字列をUTF16LE文字列に変換
// UTF16LE文字列のバイト数を返す
int convUCS4toUTF16LE(id3_ucs4_t *ucs4, char *utf16lestr, int bufsize){
	int bytes = 0;
	
	while (*ucs4 != 0 && bufsize > 2){
		*utf16lestr++ = (*ucs4 & 0x00FF);
		bufsize--;
		*utf16lestr++ = ((*ucs4 >> 8) & 0x00FF);
		bufsize--;
		ucs4++;
		bytes += 2;
	}
	*utf16lestr++ = 0;
	*utf16lestr++ = 0;
	
	return bytes;
}

// ucs4文字列をASCII文字列に変換
void convUCS4toASCII(id3_ucs4_t *ucs4, char *asciistr, int bufsize){
	while (*ucs4 != 0 && bufsize > 1){
		*asciistr++ = (*ucs4 & 0x00FF);
		bufsize--;
		ucs4++;
	}
	*asciistr++ = 0;
}


// メインメニュー／固定メッセージのイメージデータ作成
void CreateMenuData13(){
	char	menustring[ML_TITLENAME_LEN + 1];	// 表示名
	unsigned short	line1[LCD_WIDTH];
	unsigned short	line2[LCD_WIDTH];
	FILE	*fout;
	int		bytes;
	
	char	*menu1 = "フォルダ            ";
	char	*menu2 = "インデックス        ";
	char	*menu3 = "ランダム再生        ";
	char	*menu4 = "停止                ";
	char	*menu5 = "ファイル構成情報作成";
	char	*menu6 = "ステレオ／モノラル  ";
	char	*menu7 = "モノラルモード      ";
	char	*menu8 = "ステレオモード      ";
	char	*menu9 = "                    ";
	unsigned short menu_items = 6;		// メインメニューの項目数
	
	fout = fopen(MENUDAT13, "w");
	
	fwrite(&menu_items, 2, 1, fout);
	
	bytes = iconvUTF8toSJIS(menu1, strlen(menu1), menustring, ML_TITLENAME_LEN);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, menustring, bytes);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	
	bytes = iconvUTF8toSJIS(menu2, strlen(menu2), menustring, ML_TITLENAME_LEN);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, menustring, bytes);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	
	bytes = iconvUTF8toSJIS(menu3, strlen(menu3), menustring, ML_TITLENAME_LEN);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, menustring, bytes);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	
	bytes = iconvUTF8toSJIS(menu4, strlen(menu4), menustring, ML_TITLENAME_LEN);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, menustring, bytes);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	
	bytes = iconvUTF8toSJIS(menu5, strlen(menu5), menustring, ML_TITLENAME_LEN);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, menustring, bytes);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	
	bytes = iconvUTF8toSJIS(menu6, strlen(menu6), menustring, ML_TITLENAME_LEN);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, menustring, bytes);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	
	bytes = iconvUTF8toSJIS(menu7, strlen(menu7), menustring, ML_TITLENAME_LEN);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, menustring, bytes);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	
	bytes = iconvUTF8toSJIS(menu8, strlen(menu8), menustring, ML_TITLENAME_LEN);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, menustring, bytes);
	fwrite(line1, DL_LCDIMAGE, 1, fout);
	
//	for (int i=0; i<4; i++){
//		fwrite(line1, DL_LCDIMAGE, 1, fout);
//	}
	
	fclose(fout);
}


// 数字イメージファイル
void CreateDigitImageDat(){
	char	digitstring[32];
	unsigned short	line1[LCD_WIDTH];
	unsigned short	line2[LCD_WIDTH];
	FILE	*fout;
	int		bytes;
	
	char	*digits = "0123456789: ";
	
	fout = fopen(DIGITIMAGEDAT, "w");
	
	bytes = iconvUTF8toSJIS(digits, strlen(digits), digitstring, 32);
	memset(line1, 0, LCD_WIDTH*2);
	MakeLcdData(line1, line2, LCD_WIDTH, digitstring, bytes);
	fwrite(line1, DIGITIMAGE_LEN, 1, fout);
	
	fclose(fout);
}


// 履歴ファイル新規作成
void CreateHistFile(){
	FILE	*fout;
	uint16_t	lastdir = 0xFFFF;
	uint16_t	lasttrack = 0xFFFF;
	uint32_t	lastbyte = 0;
	uint8_t		stereostatus = 0;
	
	fout = fopen(HISTFILE, "w");
	fwrite(&lastdir, HS_LASTDIR_LEN, 1, fout);
	fwrite(&lasttrack, HS_LASTTRACK_LEN, 1, fout);
	fwrite(&lastbyte, HS_LASTTIME_LEN, 1, fout);
	fwrite(&FirstRandomplay, HS_NEXTRANDOMDIR_LEN, 1, fout);
	fwrite(&stereostatus, HS_STEREOSTATUS_LEN, 1, fout);
	fclose(fout);
}

